
<?php

/**
 * PHP 7.4 to PHP 8.2 Code Converter
 * 
 * This script helps convert PHP 7.4 code to be compatible with PHP 8.2
 */

class PHP74To82Converter
{
    private string $inputDir;
    private string $outputDir;
    private array $conversionRules;
    private array $statistics = [
        'files_processed' => 0,
        'files_modified' => 0,
        'changes_made' => 0,
    ];

    public function __construct(string $inputDir, string $outputDir)
    {
        $this->inputDir = rtrim($inputDir, '/');
        $this->outputDir = rtrim($outputDir, '/');
        $this->initConversionRules();
    }

    private function initConversionRules(): void
    {
        $this->conversionRules = [
            // Replace deprecated functions
            '/\bsplit\s*\(/i' => 'explode(',
            '/\bereg\s*\(/i' => 'preg_match(',
            '/\beregi\s*\(/i' => 'preg_match(\'i\'',

            // Update null coalescing operator for methods (PHP 8.0+)
            '/(\$[a-zA-Z0-9_]+)(\s*\-\>\s*[a-zA-Z0-9_]+\s*\(\s*.*?\s*\))(\s*\?\?)/s' => '$1$2 ?? ',

            // Replace deprecated array and string offset syntax
            '/\{\s*(\$[^}]+)\s*\}/' => '[$1]',

            // Constructor property promotion style (optional conversion)
            '/public\s+function\s+__construct\s*\(([^)]*)\)\s*\{((?:[^{}]|(?R))*)\}/s' => function ($matches) {
                $params = $matches[1];
                $body = $matches[2];
                
                // Extract property assignments from constructor body
                preg_match_all('/\$this->([a-zA-Z0-9_]+)\s*=\s*\$([a-zA-Z0-9_]+);/', $body, $assignments, PREG_SET_ORDER);
                
                $promotedParams = $params;
                $newBody = $body;
                
                foreach ($assignments as $assignment) {
                    $propName = $assignment[1];
                    $paramName = $assignment[2];
                    
                    if ($propName === $paramName) {
                        // Replace parameter with promoted property
                        $promotedParams = preg_replace('/(\$' . $paramName . ')/', 'private $1', $promotedParams);
                        // Remove the assignment from the body
                        $newBody = str_replace('$this->' . $propName . ' = $' . $paramName . ';', '', $newBody);
                    }
                }
                
                return 'public function __construct(' . $promotedParams . ') {' . $newBody . '}';
            },

            // Null safe operator (PHP 8.0+)
            '/(\$[a-zA-Z0-9_]+)(\s*\-\>\s*[a-zA-Z0-9_]+)(\s*\-\>\s*[a-zA-Z0-9_]+)/' => 
                function ($matches) {
                    return $matches[1] . '?' . $matches[2] . $matches[3];
                },
                
            // Replace required parameters after optional parameters (PHP 8.0+ compatibility)
            '/function\s+([a-zA-Z0-9_]+)\s*\(\s*([^)]*?=\s*[^),]+)(?:\s*,\s*)([^)=]*?[a-zA-Z0-9_]+\s*(?:,|\)))/s' =>
                function ($matches) {
                    $funcName = $matches[1];
                    $optionalParams = $matches[2];
                    $requiredParams = rtrim($matches[3], ',)');
                    
                    return "function {$funcName}($requiredParams, $optionalParams)";
                },
                
            // Update deprecated string concatenation and type casting (PHP 8.0+)
            '/"\s*\.\s*\$([a-zA-Z0-9_]+)\s*\.\s*"/' => '"{$${1}}"',
            
            // Replace deprecated each() function (PHP 7.2+)
            '/\beach\s*\(\s*\$([a-zA-Z0-9_]+)\s*\)/i' => 
                function ($matches) {
                    $arrayVar = $matches[1];
                    return "/* Replace 'each' with foreach loop:\nforeach ($$arrayVar as \$key => \$value) {\n    // Your code here\n} */\neach($$arrayVar)";
                },
        ];
    }

    public function convert(): void
    {
        $this->processDirectory($this->inputDir);
        $this->printStatistics();
    }

    private function processDirectory(string $dir): void
    {
        $items = scandir($dir);

        foreach ($items as $item) {
            if ($item === '.' || $item === '..') {
                continue;
            }

            $inputPath = "$dir/$item";
            $outputPath = str_replace($this->inputDir, $this->outputDir, $inputPath);

            if (is_dir($inputPath)) {
                if (!file_exists($outputPath)) {
                    mkdir($outputPath, 0755, true);
                }
                $this->processDirectory($inputPath);
            } elseif (pathinfo($inputPath, PATHINFO_EXTENSION) === 'php') {
                $this->processFile($inputPath, $outputPath);
            } else {
                // Copy non-PHP files as is
                if (!file_exists(dirname($outputPath))) {
                    mkdir(dirname($outputPath), 0755, true);
                }
                copy($inputPath, $outputPath);
            }
        }
    }

    private function processFile(string $inputPath, string $outputPath): void
    {
        $content = file_get_contents($inputPath);
        $newContent = $content;
        $changes = 0;

        foreach ($this->conversionRules as $pattern => $replacement) {
            $count = 0;
            if (is_callable($replacement)) {
                $newContent = preg_replace_callback($pattern, $replacement, $newContent, -1, $count);
            } else {
                $newContent = preg_replace($pattern, $replacement, $newContent, -1, $count);
            }
            $changes += $count;
        }

        // Add PHP 8.2 specific type declarations
        $newContent = $this->addReturnTypeDeclarations($newContent);
        
        // Add compatibility comments
        $newContent = $this->addCompatibilityComments($newContent);

        $this->statistics['files_processed']++;
        
        if ($newContent !== $content) {
            $this->statistics['files_modified']++;
            $this->statistics['changes_made'] += $changes;
            
            if (!file_exists(dirname($outputPath))) {
                mkdir(dirname($outputPath), 0755, true);
            }
            
            // Add header comment
            $newContent = "<?php\n// Upgraded from PHP 7.4 to PHP 8.2 compatibility\n// Original file: $inputPath\n// Upgraded on: " . date('Y-m-d H:i:s') . "\n\n" . ltrim($newContent, "<?php\n");
            
            file_put_contents($outputPath, $newContent);
        } else {
            // No changes needed, just copy the file
            if (!file_exists(dirname($outputPath))) {
                mkdir(dirname($outputPath), 0755, true);
            }
            copy($inputPath, $outputPath);
        }
    }
    
    private function addReturnTypeDeclarations(string $content): string
    {
        // This is a simplified approach - a real implementation would need more sophisticated parsing
        $pattern = '/function\s+([a-zA-Z0-9_]+)\s*\([^)]*\)\s*(?!:)/';
        return preg_replace_callback($pattern, function ($matches) {
            // Add a comment suggesting to add return type
            return $matches[0] . "/* Consider adding return type, e.g.: : void, : string, : ?object */ ";
        }, $content);
    }
    
    private function addCompatibilityComments(string $content): string
    {
        // Add PHP 8.2 deprecation warnings
        $deprecatedFeatures = [
            'utf8_encode' => 'Use mb_convert_encoding() instead',
            'utf8_decode' => 'Use mb_convert_encoding() instead',
            'create_function' => 'Use anonymous functions instead',
            'mb_ereg_replace_callback' => 'This function may behave differently in PHP 8.2',
            '$GLOBALS[\'var\']' => 'Dynamic properties are deprecated in PHP 8.2',
        ];
        
        foreach ($deprecatedFeatures as $feature => $suggestion) {
            if (strpos($feature, '$GLOBALS') !== false) {
                // Handle special case for dynamic properties
                $content = preg_replace_callback('/\$this->([a-zA-Z0-9_]+)\s*=/', function ($matches) {
                    $propName = $matches[1];
                    return "/* PHP 8.2: Consider declaring class property '$propName' */ " . $matches[0];
                }, $content);
            } else {
                // Handle function deprecations
                $content = str_replace($feature, "/* PHP 8.2 deprecated: $suggestion */ $feature", $content);
            }
        }
        
        return $content;
    }

    private function printStatistics(): void
    {
        echo "Conversion completed!\n";
        echo "Files processed: {$this->statistics['files_processed']}\n";
        echo "Files modified: {$this->statistics['files_modified']}\n";
        echo "Changes made: {$this->statistics['changes_made']}\n";
    }
}

// Example usage
if (isset($argv[1]) && isset($argv[2])) {
    $converter = new PHP74To82Converter($argv[1], $argv[2]);
    $converter->convert();
} else {
    echo "Usage: php converter.php <input_directory> <output_directory>\n";
    exit(1);
}